# SPDX-FileCopyrightText: 2022 The Ikarus Developers mueller@ibb.uni-stuttgart.de
# SPDX-License-Identifier: LGPL-3.0-or-later

from dune.generator.generator import SimpleGenerator
from dune.common.hashit import hashIt
import dune

def decoratePre(pre):
    def wrappedPre(*args, **kwargs):
        preamble = pre(*args, **kwargs)
        newPreamble = ""
        newPreamble += "#define DUNE_LOCALFEFUNCTIONS_USE_EIGEN 1\n"
        newPreamble += "#define EIGEN_DEFAULT_TO_ROW_MAJOR 1\n" #needed to have conforming Matrix storage between eigen and numpy otherwise references are not working
        newPreamble += preamble
        return newPreamble

    return wrappedPre


myAttributes = vars(SimpleGenerator).copy()
myAttributes["pre"] = decoratePre(myAttributes["pre"])
MySimpleGenerator = type("MySimpleGenerator", (object,), myAttributes)

def boundaryPatch(gridView,booleanVector) :
    dune.generator.addToFlags(pre="-DCMAKE_PREFIX_PATH=$ENV{CMAKE_PREFIX_PATH}")
    generator = MySimpleGenerator("BoundaryPatch", "Ikarus::Python")
    element_type = f"BoundaryPatch<{gridView.cppTypeName}>"

    includes = []
    includes += ["dune/fufem/boundarypatch.hh"]
    includes += ["ikarus/python/utils/boundarypatch.hh"]
    includes += gridView._includes
    moduleName = "boundaryPatch_" + hashIt(element_type)
    module = generator.load(
        includes=includes,
        typeName=element_type,
        moduleName=moduleName
    )
    return module.BoundaryPatch(gridView,booleanVector)



# SPDX-FileCopyrightText: 2022 The Ikarus Developers mueller@ibb.uni-stuttgart.de
# SPDX-License-Identifier: LGPL-3.0-or-later

from dune.generator.generator import SimpleGenerator
from dune.common.hashit import hashIt
from dune.functions.globalbasis import preBasisTypeName
from dune.functions import defaultGlobalBasis

def decoratePre(pre):
    def wrappedPre(*args, **kwargs):
        preamble = pre(*args, **kwargs)
        newPreamble = ""
        newPreamble += "#define DUNE_LOCALFEFUNCTIONS_USE_EIGEN 1\n"
        newPreamble += "#define EIGEN_DEFAULT_TO_ROW_MAJOR 1\n" #needed to have conforming Matrix storage between eigen and numpy otherwise references are not working
        newPreamble += preamble
        return newPreamble

    return wrappedPre


myAttributes = vars(SimpleGenerator).copy()
myAttributes["pre"] = decoratePre(myAttributes["pre"])
MySimpleGenerator = type("MySimpleGenerator", (object,), myAttributes)

def basis(gv, tree):

    generator = MySimpleGenerator("Basis", "Ikarus::Python")

    pbfName = preBasisTypeName(tree,gv.cppTypeName)
    element_type = f"Ikarus::Basis<{pbfName}>"

    includes = []
    # includes += ["ikarus/utils/basis.hh"]
    includes += ["ikarus/python/basis/basis.hh"]
    moduleName = "Basis_" + hashIt(element_type)
    module = generator.load(
            includes=includes,
            typeName=element_type,
            moduleName=moduleName
        )
    basis = defaultGlobalBasis(gv,tree)
    return module.Basis(basis)



# SPDX-FileCopyrightText: 2022 The Ikarus Developers mueller@ibb.uni-stuttgart.de
# SPDX-License-Identifier: LGPL-3.0-or-later

from dune.generator.generator import SimpleGenerator
from dune.common.hashit import hashIt
import dune

def decoratePre(pre):
    def wrappedPre(*args, **kwargs):
        preamble = pre(*args, **kwargs)
        newPreamble = ""
        newPreamble += "#define DUNE_LOCALFEFUNCTIONS_USE_EIGEN 1\n"
        newPreamble += "#define EIGEN_DEFAULT_TO_ROW_MAJOR 1\n" #needed to have conforming Matrix storage between eigen and numpy otherwise references are not working
        newPreamble += preamble
        return newPreamble

    return wrappedPre


myAttributes = vars(SimpleGenerator).copy()
myAttributes["pre"] = decoratePre(myAttributes["pre"])
MySimpleGenerator = type("MySimpleGenerator", (object,), myAttributes)

def materialConstructorDecorator(func):
    def wrapper(emodul,nu):
        generator = MySimpleGenerator(func.__name__, "Ikarus::Python")
        element_type = f"Ikarus::"+func.__name__+"<double>"

        includes = []
        includes += ["ikarus/finiteElements/mechanics/materials.hh"]
        includes += ["ikarus/python/finiteElements/materials/material.hh"]
        moduleName = func.__name__+"_" + hashIt(element_type)
        module = generator.load(
            includes=includes,
            typeName=element_type,
            moduleName=moduleName
        )

        # dispatcher = { 'NeoHooke' : NeoHooke, 'StVenantKirchhoff' : StVenantKirchhoff}


        return eval("module."+func.__name__+"(emodul,nu)")
    return wrapper


@materialConstructorDecorator
def StVenantKirchhoff(emodul,nu) :
    return materialConstructorDecorator

@materialConstructorDecorator
def NeoHooke(emodul,nu) :
    return materialConstructorDecorator



# SPDX-FileCopyrightText: 2022 The Ikarus Developers mueller@ibb.uni-stuttgart.de
# SPDX-License-Identifier: LGPL-3.0-or-later

import setpath
setpath.set_path()
import ikarus as iks
import ikarus.finite_elements
import ikarus.utils
import ikarus.assembler
import ikarus.dirichletValues
import numpy as np
import scipy as sp

import dune.grid
import dune.functions
from dune.vtk import vtkUnstructuredGridWriter, vtkWriter, RangeTypes, FieldInfo

if __name__ == "__main__":

    # help(iks)
    lowerLeft = []
    upperRight = []
    elements = []
    for i in range(2):
        lowerLeft.append(-1)
        upperRight.append(1)
        elements.append(3)

    req= ikarus.FErequirements()
    req.addAffordance(iks.ScalarAffordances.mechanicalPotentialEnergy)

    grid = dune.grid.structuredGrid(lowerLeft,upperRight,elements)
    grid. hierarchicalGrid . globalRefine(6)
    basisLagrange12 = dune.functions.defaultGlobalBasis(grid, dune.functions.Power(dune.functions.Lagrange(order=1),2))
    dirichletValuesT = iks.dirichletValues(basisLagrange12)

    basisLagrange1 = iks.basis(grid, dune.functions.Power(dune.functions.Lagrange(order=1),2))
    flatBasis = basisLagrange1.flat()
    d = np.zeros(len(flatBasis))
    d[0]=0.0

    lambdaLoad = iks.ValueWrapper(3.0)
    req.insertParameter(iks.FEParameter.loadfactor,lambdaLoad)

    assert req.getParameter(iks.FEParameter.loadfactor) == lambdaLoad
    req.insertGlobalSolution(iks.FESolutions.displacement,d)

    d2= req.getGlobalSolution(iks.FESolutions.displacement)

    assert ('{}'.format(hex(d2.__array_interface__['data'][0]))) == ('{}'.format(hex(d.__array_interface__['data'][0])))
    assert len(d2)== len(d)
    assert (d2== d).all()
    fes = []
    forces = np.zeros(8)
    stiffness = np.zeros((8,8))

    def volumeLoad(x,lambdaVal) :
        return np.array([lambdaVal*x[0]*2, 2*lambdaVal*x[1]*0])

    def neumannLoad(x,lambdaVal) :
        return np.array([lambdaVal*0, lambdaVal])


    neumannVertices = np.zeros(grid.size(2)*2, dtype=bool)
    assert len(neumannVertices)== len(flatBasis)

    flatBasis.interpolate(neumannVertices, lambda x :  True  if x[1]>0.9 else False)

    boundaryPatch = iks.utils.boundaryPatch(grid,neumannVertices)

    # the following should throw
    try:
        for e in grid.elements:
            iks.finite_elements.linearElasticElement(basisLagrange1,e,1000,0.2,volumeLoad,boundaryPatch)
        assert False
    except TypeError :
        pass

    for e in grid.elements:
        fes.append(iks.finite_elements.linearElasticElement(basisLagrange1,e,1000,0.2,volumeLoad,boundaryPatch,neumannLoad))

    fes[0].calculateVector(req,forces)
    fes[0].calculateMatrix(req,stiffness)
    fes[0].localView()

    dirichletValues = iks.dirichletValues(flatBasis)

    def fixFirstIndex(vec,globalIndex):
        vec[0]= True

    def fixAnotherVertex(vec,localIndex,localView):
        localView.index(localIndex)
        vec[1]= True

    def fixLeftHandEdge(vec,localIndex,localView,intersection):
        if (intersection.geometry.center[1]<-0.9):
            vec[localView.index(localIndex)]= True

    dirichletValues.fixBoundaryDOFs(fixFirstIndex)
    dirichletValues.fixBoundaryDOFsUsingLocalView(fixAnotherVertex)
    dirichletValues.fixBoundaryDOFsUsingLocalViewAndIntersection(fixLeftHandEdge)

    assembler = iks.assembler.sparseFlatAssembler(fes,dirichletValues)
    assemblerDense = iks.assembler.denseFlatAssembler(fes,dirichletValues)

    Msparse = assembler.getMatrix(req)
    forces = assembler.getVector(req)
    # Mdense = assemblerDense.getMatrix(req)
    # assert (Msparse == Mdense).all()
    # print(Mdense)
    # print("F:",forces)

    x = sp.sparse.linalg.spsolve(Msparse,-forces)
    # print(x)
    fx = flatBasis.asFunction(x)
    grid.plot()
    # grid.writeVTK(pointdata={"fx":fx},name="nameTest",number=1)
    writer = vtkWriter( grid, "nameTest",
                    pointData   = {( "displacement",(0,1)):fx}

                    )



# SPDX-FileCopyrightText: 2022 The Ikarus Developers mueller@ibb.uni-stuttgart.de
#
# SPDX-License-Identifier: CC-BY-SA-4.0

"These macros are taken and modified from https://github.com/autodiff/autodiff"

def define_env(env):
    """
    This is the hook for the functions

    - variables: the dictionary that contains the variables
    - macro: a decorator function, to declare a macro.
    """

    @env.macro
    def inputcode(filename, language, linenumbers=False, startline=0, endline=None):
        filename = '../' + filename  # file path must be given relative to root directory
        f = open(filename, 'r')
        if startline != 0 or endline is None:
            lines = f.readlines()
            lines = lines[startline:endline]
            text = "".join(lines)
        else:
            text = f.read()
        if linenumbers is False:
            textblock = f'```{{ .{language} .annotate}}\n{text}\n```'
        else:
            textblock = f'```{{ .{language} linenums="1" .annotate}} \n{text}\n```'
        return textblock

    @env.macro
    def inputcpp(filename,linenumbers=False, startline=0, endline=None):
        return inputcode(filename, 'cpp', linenumbers, startline, endline)
